from django.test import TestCase
from django.test.utils import override_settings

from edc_visit_schedule.visit import Crf, FormsCollection, FormsCollectionError


class TestFormsCollection(TestCase):
    def test_forms_collection_empty(self):
        crfs = []
        try:
            FormsCollection(*crfs)
        except FormsCollectionError as e:
            self.fail(f"FormsCollectionError unexpectedly raised. Got {e}")

    def test_forms_collection_none1(self):
        crfs = None
        try:
            FormsCollection(crfs)
        except FormsCollectionError as e:
            self.fail(f"FormsCollectionError unexpectedly raised. Got {e}")

    def test_forms_collection_order(self):
        crfs = []
        for i in range(0, 10):
            crfs.append(Crf(show_order=i, model="x.x"))
        try:
            FormsCollection(*crfs)
        except FormsCollectionError as e:
            self.fail(f"FormsCollectionError unexpectedly raised. Got {e}")
        crfs.append(Crf(show_order=0, model="x.x"))
        self.assertRaises(FormsCollectionError, FormsCollection, *crfs)

    @override_settings(SITE_ID=40)
    def test_forms_collection_excludes_by_site_id(self):
        crfs = []
        for i in range(0, 5):
            crfs.append(Crf(show_order=i, model=f"x.{i}"))
        for i in range(6, 11):
            crfs.append(Crf(show_order=i, model=f"x.{i}", site_ids=[10]))
        forms = FormsCollection(*crfs)
        self.assertEqual(len(forms.forms), 5)

    @override_settings(SITE_ID=40)
    def test_forms_collection_excludes_by_site_id2(self):
        crfs = []
        for i in range(0, 5):
            crfs.append(Crf(show_order=i, model=f"x.{i}"))
        for i in range(6, 11):
            crfs.append(Crf(show_order=i, model=f"x.{i}", site_ids=[40]))
        forms = FormsCollection(*crfs)
        self.assertEqual(len(forms.forms), 10)

    @override_settings(SITE_ID=40)
    def test_forms_collection_excludes_by_site_id3(self):
        crfs = []
        for i in range(0, 5):
            crfs.append(Crf(show_order=i, model=f"x.{i}"))
        for i in range(6, 11):
            crfs.append(Crf(show_order=i, model=f"x.{i}"))
        forms = FormsCollection(*crfs)
        self.assertEqual(len(forms.forms), 10)

    @override_settings(SITE_ID=40)
    def test_forms_collection_list_like_behaviour(self):
        forms = FormsCollection()
        for i in range(0, 5):
            forms.append(Crf(show_order=i, model=f"x.{i}"))
        for i in range(6, 11):
            forms.append(Crf(show_order=i, model=f"x.{i}"))
        self.assertEqual(len(forms), 10)
        forms.remove(Crf(show_order=1, model="x.1"))
        self.assertEqual(len(forms), 9)
        max_show_order = max([item.show_order for item in forms])
        forms.insert_last(Crf(show_order=1, model="x.1"))
        self.assertGreater(max([item.show_order for item in forms]), max_show_order)
        self.assertRaises(
            FormsCollectionError, forms.insert_last, Crf(show_order=1, model="x.1")
        )



from django.test import TestCase

from edc_visit_schedule.visit_schedule import (
    SchedulesCollection,
    SchedulesCollectionError,
)


class TestSchedulesCollection(TestCase):
    def setUp(self):
        class TestCollection(SchedulesCollection):
            key = "key"
            ordering_attr = "seq"

        class Obj:
            def __init__(self, key, seq):
                self.key = key
                self.seq = seq
                self.onschedule_model = "app_label.onschedule"
                self.offschedule_model = "app_label.offschedule"

            def validate(self, visit_schedule_name=None):
                return None

        self.test_collection = TestCollection()
        self.dummy_schedule = Obj("one", 1)
        self.test_collection.update({self.dummy_schedule.key: self.dummy_schedule})

    def test_get_by_model_raises(self):
        """Asserts bad model name raises."""
        obj = SchedulesCollection()
        self.assertRaises(SchedulesCollectionError, obj.get_schedule, "blah")

    def test_get_by_model(self):
        """Asserts can get the object using either model,
        onschedule_model or offschedule_model
        """
        self.assertEqual(
            self.test_collection.get_schedule(model="app_label.onschedule"),
            self.dummy_schedule,
        )
        self.assertEqual(
            self.test_collection.get_schedule("app_label.onschedule"),
            self.dummy_schedule,
        )

    def test_validate(self):
        self.test_collection.validate()



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_consent import site_consents
from edc_consent.consent import Consent
from edc_constants.constants import FEMALE, MALE
from edc_protocol import Protocol
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow

from edc_visit_schedule.models import SubjectScheduleHistory
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.subject_schedule import SubjectSchedule, SubjectScheduleError
from edc_visit_schedule.visit_schedule import VisitSchedule
from visit_schedule_app.models import OffSchedule, OnSchedule, SubjectConsent


class TestSubjectSchedule(SiteTestCaseMixin, TestCase):
    def setUp(self):
        v1_consent = Consent(
            "visit_schedule_app.subjectconsent",
            version="1",
            start=Protocol().study_open_datetime,
            end=Protocol().study_close_datetime,
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )
        site_consents.registry = {}
        site_consents.register(v1_consent)
        site_visit_schedules._registry = {}
        self.visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.SubjectOffstudy",
            death_report_model="visit_schedule_app.DeathReport",
        )

        self.schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.OnSchedule",
            offschedule_model="visit_schedule_app.OffSchedule",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
        )
        self.schedule3 = Schedule(
            name="schedule_three",
            onschedule_model="visit_schedule_app.OnScheduleThree",
            offschedule_model="visit_schedule_app.OffScheduleThree",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
        )

        self.visit_schedule.add_schedule(self.schedule)
        self.visit_schedule.add_schedule(self.schedule3)
        site_visit_schedules.register(self.visit_schedule)

        self.visit_schedule_two = VisitSchedule(
            name="visit_schedule_two",
            verbose_name="Visit Schedule Two",
            offstudy_model="visit_schedule_app.SubjectOffstudy",
            death_report_model="visit_schedule_app.DeathReport",
        )

        self.schedule_two_1 = Schedule(
            name="schedule_two",
            onschedule_model="visit_schedule_app.OnScheduleTwo",
            offschedule_model="visit_schedule_app.OffScheduleTwo",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
        )
        self.schedule_two_2 = Schedule(
            name="schedule_four",
            onschedule_model="visit_schedule_app.OnScheduleFour",
            offschedule_model="visit_schedule_app.OffScheduleFour",
            appointment_model="edc_appointment.appointment",
            consent_model="visit_schedule_app.subjectconsent",
        )

        self.visit_schedule_two.add_schedule(self.schedule_two_1)
        self.visit_schedule_two.add_schedule(self.schedule_two_2)
        site_visit_schedules.register(self.visit_schedule_two)
        self.subject_identifier = "111111"
        SubjectConsent.objects.create(subject_identifier=self.subject_identifier)

    def test_onschedule_updates_history(self):
        """Asserts returns the correct instances for the schedule."""
        for onschedule_model, schedule_name in [
            ("visit_schedule_app.onscheduletwo", "schedule_two"),
            ("visit_schedule_app.onschedulefour", "schedule_four"),
        ]:
            with self.subTest(onschedule_model=onschedule_model, schedule_name=schedule_name):
                visit_schedule, schedule = site_visit_schedules.get_by_onschedule_model(
                    onschedule_model
                )
                subject_schedule = SubjectSchedule(
                    visit_schedule=visit_schedule, schedule=schedule
                )
                subject_schedule.put_on_schedule(
                    subject_identifier=self.subject_identifier,
                    onschedule_datetime=get_utcnow(),
                )
                try:
                    SubjectScheduleHistory.objects.get(
                        subject_identifier=self.subject_identifier,
                        schedule_name=schedule_name,
                    )
                except ObjectDoesNotExist:
                    self.fail("ObjectDoesNotExist unexpectedly raised")

    def test_multpile_consents(self):
        """Asserts does not raise if more than one consent
        for this subject
        """
        subject_identifier = "ABCDEF"
        SubjectConsent.objects.create(subject_identifier=subject_identifier, version="1")
        SubjectConsent.objects.create(subject_identifier=subject_identifier, version="2")
        visit_schedule, schedule = site_visit_schedules.get_by_onschedule_model(
            "visit_schedule_app.onscheduletwo"
        )
        subject_schedule = SubjectSchedule(visit_schedule=visit_schedule, schedule=schedule)
        try:
            subject_schedule.put_on_schedule(
                subject_identifier=subject_identifier, onschedule_datetime=get_utcnow()
            )
        except SubjectScheduleError:
            self.fail("SubjectScheduleError unexpectedly raised.")

    def test_resave(self):
        """Asserts returns the correct instances for the schedule."""
        visit_schedule, schedule = site_visit_schedules.get_by_onschedule_model(
            "visit_schedule_app.onscheduletwo"
        )
        subject_schedule = SubjectSchedule(visit_schedule=visit_schedule, schedule=schedule)
        subject_schedule.put_on_schedule(
            subject_identifier=self.subject_identifier, onschedule_datetime=get_utcnow()
        )
        subject_schedule.resave(subject_identifier=self.subject_identifier)

    def test_put_on_schedule(self):
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            "visit_schedule_app.onschedule"
        )
        self.assertRaises(
            ObjectDoesNotExist,
            OnSchedule.objects.get,
            subject_identifier=self.subject_identifier,
        )
        schedule.put_on_schedule(
            subject_identifier=self.subject_identifier, onschedule_datetime=get_utcnow()
        )
        try:
            OnSchedule.objects.get(subject_identifier=self.subject_identifier)
        except ObjectDoesNotExist:
            self.fail("ObjectDoesNotExist unexpectedly raised")

    def test_take_off_schedule(self):
        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name="visit_schedule"
        )
        schedule = visit_schedule.schedules.get("schedule")
        schedule.put_on_schedule(subject_identifier=self.subject_identifier)
        schedule.take_off_schedule(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=get_utcnow(),
        )
        try:
            OffSchedule.objects.get(subject_identifier=self.subject_identifier)
        except ObjectDoesNotExist:
            self.fail("ObjectDoesNotExist unexpectedly raised")



from django.db import models
from edc_model.models import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_utils import get_utcnow

from ..site_visit_schedules import site_visit_schedules
from .schedule_model_mixin import ScheduleModelMixin


class OnScheduleModelMixin(ScheduleModelMixin):
    """A model mixin for a schedule's onschedule model."""

    onschedule_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=get_utcnow,
    )

    def save(self, *args, **kwargs):
        self.report_datetime = self.onschedule_datetime
        super().save(*args, **kwargs)

    def put_on_schedule(self):
        _, schedule = site_visit_schedules.get_by_onschedule_model(self._meta.label_lower)
        schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=self.onschedule_datetime,
        )

    @property
    def visit_schedule(self):
        """Returns a visit schedule object."""
        return site_visit_schedules.get_by_onschedule_model(
            onschedule_model=self._meta.label_lower
        )[0]

    @property
    def schedule(self):
        """Returns a schedule object."""
        return site_visit_schedules.get_by_onschedule_model(
            onschedule_model=self._meta.label_lower
        )[1]

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=["id", "subject_identifier", "onschedule_datetime", "site"])
        ]



from django.db import models

from ..subject_schedule import SubjectSchedule


class SubjectScheduleModelMixin(models.Model):
    """A mixin for CRF models to add the ability to determine
    if the subject is on/off schedule.
    """

    # If True, compares report_datetime and offschedule_datetime as datetimes
    # If False, (Default) compares report_datetime and
    # offschedule_datetime as dates
    offschedule_compare_dates_as_datetimes = False
    subject_schedule_cls = SubjectSchedule

    def validate_subject_schedule_status(self):
        visit_schedule = self.visit.appointment.visit_schedule
        schedule = self.visit.appointment.schedule
        subject_identifier = self.visit.subject_identifier
        subject_schedule = self.subject_schedule_cls(
            visit_schedule=visit_schedule, schedule=schedule
        )
        subject_schedule.onschedule_or_raise(
            subject_identifier=subject_identifier,
            report_datetime=self.visit.report_datetime,
            compare_as_datetimes=self.offschedule_compare_dates_as_datetimes,
        )

    def save(self, *args, **kwargs):
        self.validate_subject_schedule_status()
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



from django.core.exceptions import ObjectDoesNotExist
from django.views.generic.base import ContextMixin
from edc_utils import get_utcnow

from .site_visit_schedules import site_visit_schedules


class VisitScheduleViewMixin(ContextMixin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.onschedule_models = []
        self.current_schedule = None
        self.current_visit_schedule = None
        self.current_onschedule_model = None
        self.visit_schedules = {}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        for visit_schedule in site_visit_schedules.visit_schedules.values():
            for schedule in visit_schedule.schedules.values():
                try:
                    onschedule_model_obj = schedule.onschedule_model_cls.objects.get(
                        subject_identifier=self.subject_identifier
                    )
                except ObjectDoesNotExist:
                    pass
                else:
                    if schedule.is_onschedule(
                        subject_identifier=self.kwargs.get("subject_identifier"),
                        report_datetime=get_utcnow(),
                    ):
                        self.current_schedule = schedule
                        self.current_visit_schedule = visit_schedule
                        self.current_onschedule_model = onschedule_model_obj
                    self.onschedule_models.append(onschedule_model_obj)
                    self.visit_schedules.update({visit_schedule.name: visit_schedule})

        context.update(
            visit_schedules=self.visit_schedules,
            current_onschedule_model=self.current_onschedule_model,
            onschedule_models=self.onschedule_models,
            current_schedule=self.current_schedule,
            current_visit_schedule=self.current_visit_schedule,
        )
        return context



from .off_schedule_model_mixin import OffScheduleModelMixin
from .on_schedule_model_mixin import OnScheduleModelMixin
from .schedule_model_mixin import CurrentSiteManager
from .subject_on_schedule_model_mixin import SubjectOnScheduleModelMixin
from .subject_schedule_model_mixin import SubjectScheduleModelMixin
from .visit_schedule_model_mixins import (
    VisitCodeFieldsModelMixin,
    VisitScheduleFieldsModelMixin,
    VisitScheduleMethodsError,
    VisitScheduleMethodsModelMixin,
    VisitScheduleModelMixin,
)



from typing import Protocol

from django.db import models

from ..schedule import Schedule, VisitCollection
from ..site_visit_schedules import site_visit_schedules
from ..visit import Visit
from ..visit_schedule import VisitSchedule


class MetaStub(Protocol):
    visit_schedule_name: str
    schedule_name: str
    ...


class VisitScheduleStub(Protocol):
    visit_schedule: VisitSchedule
    schedule: Schedule
    visit_schedule_name: str
    schedule_name: str
    visit_code: str
    visit_code_sequence: int
    _meta: MetaStub
    ...


class VisitScheduleMethodsError(Exception):
    pass


class VisitScheduleModelMixinError(Exception):
    pass


class VisitScheduleMethodsModelMixin(models.Model):
    """A model mixin that adds methods used to work with the visit schedule.

    Declare with VisitScheduleFieldsModelMixin or the fields from
    VisitScheduleFieldsModelMixin
    """

    @property
    def visit(self) -> Visit:
        """Returns the visit object from the schedule object
        for this visit code.

        Note: This is not a model instance.
        """
        return self.visit_from_schedule

    @property
    def visit_from_schedule(self: VisitScheduleStub) -> Visit:
        """Returns the visit object from the schedule object
        for this visit code.

        Note: This is not a model instance.
        """
        visit = self.schedule.visits.get(self.visit_code)
        if not visit:
            raise VisitScheduleModelMixinError(
                f"Visit not found in schedule. Expected one of {self.schedule.visits}. "
                f"Got {self.visit_code}."
            )
        return visit

    @property
    def visits(self: VisitScheduleStub) -> VisitCollection:
        """Returns all visit objects from the schedule object."""
        return self.schedule.visits

    @property
    def schedule(self: VisitScheduleStub) -> Schedule:
        """Returns a schedule object from Meta.visit_schedule_name or
        self.schedule_name.

        Declared on Meta like this:
            visit_schedule_name = 'visit_schedule_name.schedule_name'
        """
        return self.visit_schedule.schedules.get(self.schedule_name)

    @property
    def visit_schedule(self: VisitScheduleStub) -> VisitSchedule:
        """Returns a visit schedule object from Meta.visit_schedule_name.

        Declared on Meta like this:
            visit_schedule_name = 'my_visit_schedule_name.my_schedule_name'

        or, for example, in the case of an offstudy model:
            visit_schedule_name = 'my_visit_schedule_name'
        """
        try:
            visit_schedule_name, _ = self._meta.visit_schedule_name.split(".")
        except ValueError:
            visit_schedule_name = self._meta.visit_schedule_name
        except AttributeError:
            visit_schedule_name = self.visit_schedule_name
        return site_visit_schedules.get_visit_schedule(visit_schedule_name=visit_schedule_name)

    class Meta:
        abstract = True


class VisitScheduleFieldsModelMixin(models.Model):
    """A model mixin that adds fields required to work with the visit
    schedule methods on the VisitScheduleMethodsModelMixin.

    Note: visit_code is not included.
    """

    visit_schedule_name = models.CharField(
        max_length=25,
        editable=False,
        help_text='the name of the visit schedule used to find the "schedule"',
    )

    schedule_name = models.CharField(max_length=25, editable=False)

    class Meta:
        abstract = True


class VisitCodeFieldsModelMixin(models.Model):
    visit_code = models.CharField(max_length=25, null=True, editable=False)

    visit_code_sequence = models.IntegerField(
        verbose_name="Sequence",
        default=0,
        null=True,
        blank=True,
        help_text=(
            "An integer to represent the sequence of additional "
            "appointments relative to the base appointment, 0, needed "
            "to complete data collection for the timepoint. (NNNN.0)"
        ),
    )

    class Meta:
        abstract = True


class VisitScheduleModelMixin(
    VisitScheduleFieldsModelMixin,
    VisitCodeFieldsModelMixin,
    VisitScheduleMethodsModelMixin,
    models.Model,
):

    """A model mixin that adds adds field attributes and methods that
    link a model instance to its schedule.

    This mixin is used with Appointment and Visit models via their
    respective model mixins.
    """

    class Meta:
        abstract = True



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from ..constants import ON_SCHEDULE
from ..model_mixins import OffScheduleModelMixin
from ..site_visit_schedules import SiteVisitScheduleError, site_visit_schedules


@receiver(post_save, weak=False, dispatch_uid="offschedule_model_on_post_save")
def offschedule_model_on_post_save(sender, instance, raw, update_fields, **kwargs):
    if not raw and not update_fields:
        if issubclass(sender, (OffScheduleModelMixin,)):
            _, schedule = site_visit_schedules.get_by_offschedule_model(
                instance._meta.label_lower
            )
            schedule.take_off_schedule(
                subject_identifier=instance.subject_identifier,
                offschedule_datetime=instance.offschedule_datetime,
            )


@receiver(post_save, weak=False, dispatch_uid="onschedule_model_on_post_save")
def onschedule_model_on_post_save(instance, raw, update_fields, **kwargs):
    if not raw and not update_fields:
        try:
            instance.put_on_schedule()
        except AttributeError:
            pass


@receiver(post_delete, weak=False, dispatch_uid="offschedule_model_on_post_delete")
def offschedule_model_on_post_delete(instance, **kwargs):
    try:
        _, schedule = site_visit_schedules.get_by_offschedule_model(instance._meta.label_lower)
    except SiteVisitScheduleError:
        pass
    else:
        if schedule.offschedule_model == instance._meta.label_lower:
            history_obj = schedule.history_model_cls.objects.get(
                subject_identifier=instance.subject_identifier,
                onschedule_model=schedule.onschedule_model,
            )
            history_obj.offschedule_datetime = None
            history_obj.schedule_status = ON_SCHEDULE
            history_obj.save()
            onschedule_model_obj = schedule.onschedule_model_cls.objects.get(
                subject_identifier=instance.subject_identifier
            )
            onschedule_model_obj.save()


@receiver(post_delete, weak=False, dispatch_uid="onschedule_model_on_post_delete")
def onschedule_model_on_post_delete(instance, **kwargs):
    try:
        _, schedule = site_visit_schedules.get_by_offschedule_model(instance._meta.label_lower)
    except SiteVisitScheduleError:
        pass
    else:
        if schedule.onschedule_model == instance._meta.label_lower:
            schedule.history_model_cls.objects.filter(
                subject_identifier=instance.subject_identifier
            ).delete()


@receiver(post_save, weak=False, dispatch_uid="put_subject_on_schedule_on_post_save")
def put_subject_on_schedule_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        try:
            instance.put_subject_on_schedule_on_post_save(created)
        except AttributeError as e:
            if "put_subject_on_schedule_on_post_save" not in str(e):
                raise



import itertools
from collections import OrderedDict


class OrderedCollection(OrderedDict):

    key: str = None  # key name in dictionary key/value pair
    ordering_attr: str = None  # value.attrname to order dictionary on.

    def update(self, *args, **kwargs) -> None:
        """Updates and reorders."""

        def key_order(v):
            return getattr(v, self.ordering_attr)

        super().update(*args, **kwargs)
        od = self.copy()
        self.clear()
        super().update(**{getattr(v, self.key): v for v in sorted(od.values(), key=key_order)})

    @property
    def first(self):
        """Returns the first item."""
        return next(iter(self.values()))

    @property
    def last(self):
        """Returns the last item."""
        return next(reversed(self.values()))

    def previous(self, key):
        """Returns the previous item or None."""
        return self.get(self._iter_keys(key=key, reverse=True))

    def next(self, key):
        """Returns the next item or None."""
        return self.get(self._iter_keys(key=key))

    def _iter_keys(self, key=None, reverse=None):
        if reverse:
            seq = reversed(self.keys())
        else:
            seq = iter(self.keys())
        keys = itertools.dropwhile(lambda x: x != key, seq)
        try:
            k = next(keys)
        except StopIteration:
            pass
        else:
            assert k == key
        try:
            return next(keys)
        except StopIteration:
            return None



from django.core.checks import Warning

from .site_visit_schedules import site_visit_schedules


def visit_schedule_check(app_configs, **kwargs):
    errors = []

    if not site_visit_schedules.visit_schedules:
        errors.append(
            Warning("No visit schedules have been registered!", id="edc_visit_schedule.001")
        )
    site_results = site_visit_schedules.check()
    for key, results in site_results.items():
        for result in results:
            errors.append(Warning(result, id=f"edc_visit_schedule.{key}"))
    return errors



from django import forms
from django.apps import apps as django_apps
from edc_utils import formatted_datetime

from .constants import DAY1
from .site_visit_schedules import site_visit_schedules


class OnScheduleError(Exception):
    pass


def is_baseline(subject_visit) -> bool:
    return (
        subject_visit.appointment.visit_code == DAY1
        and subject_visit.appointment.visit_code_sequence == 0
    )


def raise_if_baseline(subject_visit) -> None:
    if subject_visit and is_baseline(subject_visit=subject_visit):
        raise forms.ValidationError("This form is not available for completion at baseline.")


def raise_if_not_baseline(subject_visit) -> None:
    if subject_visit and not is_baseline(subject_visit=subject_visit):
        raise forms.ValidationError("This form is only available for completion at baseline.")


def get_onschedule_models(subject_identifier=None, report_datetime=None):
    """Returns a list of onschedule models, in label_lower format,
    for this subject and date.
    """
    onschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    for onschedule_model_obj in subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    ):
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        onschedule_models.append(schedule.onschedule_model)
    return onschedule_models


def get_offschedule_models(subject_identifier=None, report_datetime=None):
    """Returns a list of offschedule models, in label_lower format,
    for this subject and date.

    Subject status must be ON_SCHEDULE.

    See also, manager method `onschedules`.
    """
    offschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    onschedule_models = subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    )
    for onschedule_model_obj in onschedule_models:
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        offschedule_models.append(schedule.offschedule_model)
    return offschedule_models


def off_schedule_or_raise(
    subject_identifier=None,
    report_datetime=None,
    visit_schedule_name=None,
    schedule_name=None,
):
    """Returns True if subject is on the given schedule
    on this date.
    """
    visit_schedule = site_visit_schedules.get_visit_schedule(
        visit_schedule_name=visit_schedule_name
    )
    schedule = visit_schedule.schedules.get(schedule_name)
    if schedule.is_onschedule(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    ):
        raise OnScheduleError(
            f"Not allowed. Subject {subject_identifier} is on schedule "
            f"{visit_schedule.verbose_name}.{schedule_name} on "
            f"{formatted_datetime(report_datetime)}. "
            f"See model '{schedule.offschedule_model_cls().verbose_name}'"
        )



from django import forms

from ..subject_schedule import NotOnScheduleError, NotOnScheduleForDateError


class SubjectScheduleModelFormMixin:
    """A ModelForm mixin to validate that the subject is onschedule.

    Declare with `forms.ModelForm`.

    See `self._meta.model.visit_model_attr()`.
    """

    def clean(self):
        cleaned_data = super().clean()
        self.validate_subject_schedule()
        return cleaned_data

    def validate_subject_schedule(self) -> None:
        subject_visit = self.cleaned_data.get(self._meta.model.visit_model_attr())
        if subject_visit:
            visit_schedule = subject_visit.appointment.visit_schedule
            schedule = subject_visit.appointment.schedule
            subject_schedule = self.get_subject_schedule(visit_schedule, schedule)
            try:
                subject_schedule.onschedule_or_raise(
                    subject_identifier=subject_visit.subject_identifier,
                    report_datetime=subject_visit.report_datetime,
                    compare_as_datetimes=(
                        self._meta.model.offschedule_compare_dates_as_datetimes
                    ),
                )
            except (NotOnScheduleError, NotOnScheduleForDateError) as e:
                raise forms.ValidationError(str(e))

    def get_subject_schedule(self, visit_schedule, schedule):
        try:
            subject_schedule = self._meta.model.subject_schedule_cls(
                visit_schedule=visit_schedule, schedule=schedule
            )
        except AttributeError as e:
            if "subject_schedule_cls" in str(e):
                raise AttributeError(
                    f"Was model `{self._meta.model._meta.label_lower}` declared "
                    f"with `SubjectScheduleModelMixin`? Got {e}"
                )
            else:
                raise
        return subject_schedule



from collections import OrderedDict
from datetime import datetime
from typing import Any

from ..ordered_collection import OrderedCollection


class VisitCollectionError(Exception):
    pass


class VisitCollection(OrderedCollection):

    key: str = "code"
    ordering_attr: str = "timepoint"

    def __get__(self, instance: Any, owner: Any):
        value = super().__get__(instance, owner)  # type: ignore
        if value is None:
            raise VisitCollectionError(f"Unknown visit. Got {instance}")
        return value

    def timepoint_dates(self, dt: datetime) -> dict:
        """Returns an ordered dictionary of visit dates calculated
        relative to the first visit.
        """
        timepoint_dates = OrderedDict()
        for visit in self.values():
            try:
                timepoint_datetime = dt + visit.rbase
            except TypeError as e:
                raise VisitCollectionError(
                    f"Invalid visit.rbase. visit.rbase={visit.rbase}. "
                    f"See {repr(visit)}. Got {e}."
                )
            else:
                visit.timepoint_datetime = timepoint_datetime
            timepoint_dates.update({visit: visit.timepoint_datetime})
        return timepoint_dates



# Generated by Django 3.0.4 on 2020-05-12 21:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_visit_schedule", "0007_auto_20190922_0439"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="subjectschedulehistory",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
            },
        ),
        migrations.AlterModelOptions(
            name="visitschedule",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("visit_schedule_name", "schedule_name", "visit_code"),
            },
        ),
    ]



visit_schedule_fields = ["visit_schedule_name", "schedule_name", "visit_code"]

visit_schedule_fieldset_tuple = (
    "Visit Schedule",
    {"classes": ("collapse",), "fields": visit_schedule_fields},
)

visit_schedule_only_fields = ("visit_schedule_name", "schedule_name")

visit_schedule_only_fieldset_tuple = (
    "Visit Schedule",
    {"classes": ("collapse",), "fields": visit_schedule_only_fields},
)



from django.forms import ModelForm, ValidationError
from edc_utils import get_utcnow

from ..site_visit_schedules import site_visit_schedules
from ..subject_schedule import InvalidOffscheduleDate


class OffScheduleModelFormMixin(ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        subject_identifier = cleaned_data.get("subject_identifier")

        offschedule_datetime = (
            cleaned_data.get(self.offschedule_datetime_field) or get_utcnow()
        )
        visit_schedule, schedule = site_visit_schedules.get_by_offschedule_model(
            self._meta.model._meta.label_lower
        )
        history_obj = schedule.history_model_cls.objects.get(
            subject_identifier=subject_identifier,
            schedule_name=schedule.name,
            visit_schedule_name=visit_schedule.name,
        )
        try:
            schedule.subject.update_history_or_raise(
                history_obj=history_obj,
                subject_identifier=subject_identifier,
                offschedule_datetime=offschedule_datetime,
                update=False,
            )
        except InvalidOffscheduleDate as e:
            raise ValidationError(e)
        self.validate_visit_tracking_reports()
        return cleaned_data

    @property
    def offschedule_datetime_field(self):

        try:
            offschedule_datetime_field = self._meta.model.offschedule_datetime_field
        except AttributeError:
            offschedule_datetime_field = "offschedule_datetime"
        return offschedule_datetime_field

    # TODO: validate_visit_tracking_reports before taking off schedule
    def validate_visit_tracking_reports(self):
        """Asserts that all visit tracking reports
        have been submitted.
        """
        pass



from .crf import Crf


class RequisitionError(Exception):
    pass


class RequisitionLookupError(Exception):
    pass


class ScheduledRequisitionError(Exception):
    pass


class Requisition(Crf):
    def __init__(self, panel=None, required: bool = None, **kwargs):
        required = False if required is None else required
        self.panel = panel
        if not self.panel.requisition_model:
            raise RequisitionError(
                f"Invalid requisition model. Got None. "
                f"See {repr(panel)}. "
                f"Was the panel referred to by this schedule's requisition "
                f"added to a lab profile and registered with site_labs?"
            )
        super().__init__(required=required, model=self.panel.requisition_model, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.show_order}, {self.panel.name})"

    def __str__(self):
        required = "Required" if self.required else ""
        return f"{self.panel.name} {required}"

    @property
    def name(self):
        return self.panel.name

    @property
    def verbose_name(self):
        return self.panel.verbose_name

    def validate(self):
        """Raises an exception if a Requisition model lookup fails
        or if a panel is referred to that is not known to any
        lab_profiles

        See also: edc_lab.
        """
        from edc_lab.site_labs import site_labs

        try:
            self.panel.requisition_model_cls
        except LookupError as e:
            raise RequisitionLookupError(e) from e

        for lab_profile in site_labs.registry.values():
            try:
                panels = self.panel.panels
            except AttributeError:
                panels = [self.panel]
            for panel in panels:
                if panel.name not in lab_profile.panels:
                    raise ScheduledRequisitionError(
                        f"Panel does not exist in lab profiles. " f"Got {repr(panel)}"
                    )



from datetime import date

from django.db import models
from django.db.models.deletion import PROTECT
from edc_appointment.models import Appointment
from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_crf.crf_model_mixin import CrfModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel
from edc_offstudy.model_mixins import OffstudyModelManager, OffstudyModelMixin
from edc_reference.model_mixins import ReferenceModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_tracking.model_mixins import VisitModelMixin

from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin


class SubjectVisit(
    VisitModelMixin,
    ReferenceModelMixin,
    CreatesMetadataModelMixin,
    SiteModelMixin,
    RequiresConsentFieldsModelMixin,
    BaseUuidModel,
):
    appointment = models.OneToOneField(
        Appointment, on_delete=PROTECT, related_name="test_visit_schedule_appointment"
    )

    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()

    reason = models.CharField(max_length=25, null=True)


class SubjectConsent(
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectOffstudy(OffstudyModelMixin, BaseUuidModel):
    objects = OffstudyModelManager()


class SubjectOffstudyFive(OffstudyModelMixin, BaseUuidModel):
    objects = OffstudyModelManager()


class SubjectOffstudySix(OffstudyModelMixin, BaseUuidModel):
    objects = OffstudyModelManager()


class SubjectOffstudySeven(OffstudyModelMixin, BaseUuidModel):
    objects = OffstudyModelManager()


class DeathReport(BaseUuidModel):
    subject_identifier = models.CharField(max_length=25, null=True)

    report_datetime = models.DateTimeField()


# visit_schedule


class OnSchedule(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffSchedule(OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


class OnScheduleThree(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleThree(OffScheduleModelMixin, BaseUuidModel):
    class Meta(OffScheduleModelMixin.Meta):
        pass


# visit_schedule_two


class OnScheduleTwo(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleTwo(OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleFour(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleFour(OffScheduleModelMixin, BaseUuidModel):
    pass


class OnScheduleFive(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleFive(OffScheduleModelMixin, BaseUuidModel):
    my_offschedule_datetime = models.DateTimeField()

    class Meta(OffScheduleModelMixin.Meta):
        offschedule_datetime_field = "my_offschedule_datetime"


class OnScheduleSix(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleSix(OffScheduleModelMixin, BaseUuidModel):
    my_offschedule_date = models.DateField()

    class Meta(OffScheduleModelMixin.Meta):
        offschedule_datetime_field = "my_offschedule_date"


class BadOffSchedule1(OffScheduleModelMixin, BaseUuidModel):
    """Meta.OffScheduleModelMixin.offschedule_datetime_field
    is None"""

    my_offschedule_date = models.DateField()

    class Meta(OffScheduleModelMixin.Meta):
        offschedule_datetime_field = None


class OnScheduleSeven(OnScheduleModelMixin, BaseUuidModel):
    pass


class OffScheduleSeven(OffScheduleModelMixin, BaseUuidModel):
    """Is Missing Meta.OffScheduleModelMixin"""

    class Meta:
        pass


# class CrfOne(VisitTrackingCrfModelMixin, BaseUuidModel):
class CrfOne(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    f1 = models.CharField(max_length=50, null=True, blank=True)

    f2 = models.CharField(max_length=50, null=True, blank=True)

    f3 = models.CharField(max_length=50, null=True, blank=True)



import re

from django.core.management.color import color_style

from ..site_visit_schedules import SiteVisitScheduleError, site_visit_schedules
from ..subject_schedule import (
    NotOnScheduleError,
    NotOnScheduleForDateError,
    SubjectSchedule,
    SubjectScheduleError,
)
from ..visit import Visit
from .visit_collection import VisitCollection
from .window import Window

style = color_style()


class ScheduleNameError(Exception):
    pass


class AlreadyRegisteredVisit(Exception):
    pass


class Schedule:
    """A class that represents a "schedule" of visits.

    Is contained by a "visit schedule".

    Contains an ordered dictionary of visit instances and the onschedule
    and offschedule models used to get on and off the schedule.
    """

    name_regex = r"[a-z0-9\_\-]+$"
    visit_cls = Visit
    visit_collection_cls = VisitCollection
    subject_schedule_cls = SubjectSchedule
    window_cls = Window

    def __init__(
        self,
        name=None,
        verbose_name=None,
        sequence=None,
        onschedule_model=None,
        offschedule_model=None,
        loss_to_followup_model=None,
        appointment_model=None,
        consent_model=None,
    ):
        self._subject = None
        self.visits = self.visit_collection_cls()
        if not name or not re.match(r"[a-z0-9_\-]+$", name):
            raise ScheduleNameError(
                f"Invalid name. Got '{name}'. May only contains numbers, "
                "lower case letters and '_'."
            )
        else:
            self.name = name
        self.verbose_name = verbose_name or name
        self.sequence = sequence or name

        self.appointment_model = appointment_model.lower()
        self.consent_model = consent_model.lower()
        self.offschedule_model = offschedule_model.lower()
        self.onschedule_model = onschedule_model.lower()
        if loss_to_followup_model:
            self.loss_to_followup_model = loss_to_followup_model.lower()

    def check(self):
        warnings = []
        try:
            self.subject.check()
        except (SiteVisitScheduleError, SubjectScheduleError) as e:
            warnings.append(f"{e} See schedule '{self.name}'.")
        return warnings

    def __repr__(self):
        return f"Schedule({self.name})"

    def __str__(self):
        return self.name

    def add_visit(self, visit=None, **kwargs):
        """Adds a unique visit to the schedule."""
        visit = visit or self.visit_cls(**kwargs)
        for attr in ["code", "title", "timepoint", "rbase"]:
            if getattr(visit, attr) in [getattr(v, attr) for v in self.visits.values()]:
                raise AlreadyRegisteredVisit(
                    f"Visit already registered. Got visit={visit} "
                    f"(offending attr='{attr}'). "
                    f"See schedule '{self}'"
                )
        self.visits.update({visit.code: visit})
        return visit

    @property
    def field_value(self):
        return self.name

    def crf_required_at(self, label_lower: str) -> list:
        """Returns a list of visit codes where the CRF is required
        by default.
        """
        visit_codes = []
        for visit_code, visit in self.visits.items():
            if label_lower in [form.model for form in visit.crfs if form.required]:
                visit_codes.append(visit_code)
        return visit_codes

    def requisition_required_at(self, requisition_panel) -> list:
        """Returns a list of visit codes where the requisition is
        required by default.

        A requisition is found by its panel.
        """
        visit_codes = []
        for visit_code, visit in self.visits.items():
            if requisition_panel in [
                form.panel for form in visit.requisitions if form.required
            ]:
                visit_codes.append(visit_code)
        return visit_codes

    @property
    def subject(self):
        """Returns a SubjectSchedule instance.

        Note: SubjectSchedule puts a subject on to a schedule or takes a subject
        off of a schedule.
        """
        if not self._subject:
            visit_schedule, schedule = site_visit_schedules.get_by_onschedule_model(
                self.onschedule_model
            )
            if schedule.name != self.name:
                raise ValueError(
                    f"Site visit schedules return the wrong schedule object. "
                    f"Expected {repr(self)} for onschedule_model={self.onschedule_model}. "
                    f"Got {repr(schedule)}."
                )
            self._subject = self.subject_schedule_cls(
                visit_schedule=visit_schedule, schedule=self
            )
        return self._subject

    def put_on_schedule(self, **kwargs):
        """Wrapper method to puts a subject onto this schedule."""
        self.subject.put_on_schedule(**kwargs)

    def refresh_schedule(self, **kwargs):
        """Resaves the onschedule model to, for example, refresh
        appointments.
        """
        self.subject.resave(**kwargs)

    def take_off_schedule(self, offschedule_datetime=None, **kwargs):
        self.subject.take_off_schedule(offschedule_datetime=offschedule_datetime, **kwargs)

    def is_onschedule(self, **kwargs):
        try:
            self.subject.onschedule_or_raise(compare_as_datetimes=True, **kwargs)
        except (NotOnScheduleError, NotOnScheduleForDateError):
            return False
        return True

    def datetime_in_window(self, **kwargs):
        return self.window_cls(name=self.name, visits=self.visits, **kwargs).datetime_in_window

    @property
    def onschedule_model_cls(self):
        return self.subject.onschedule_model_cls

    @property
    def offschedule_model_cls(self):
        return self.subject.offschedule_model_cls

    @property
    def history_model_cls(self):
        return self.subject.history_model_cls

    @property
    def appointment_model_cls(self):
        return self.subject.appointment_model_cls

    @property
    def visit_model_cls(self):
        return self.subject.visit_model_cls

    @property
    def consent_model_cls(self):
        return self.subject.consent_model_cls

    def to_dict(self):
        return {k: v.to_dict() for k, v in self.visits.items()}



import arrow
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import options
from edc_model.models import date_not_future, datetime_not_future
from edc_protocol.validators import (
    date_not_before_study_start,
    datetime_not_before_study_start,
)
from edc_utils import get_utcnow

from ..site_visit_schedules import site_visit_schedules
from .schedule_model_mixin import ScheduleModelMixin

if "offschedule_datetime_field" not in options.DEFAULT_NAMES:
    options.DEFAULT_NAMES = options.DEFAULT_NAMES + ("offschedule_datetime_field",)


class OffScheduleModelMixin(ScheduleModelMixin):
    """Model mixin for a schedule's OffSchedule model."""

    offschedule_datetime = models.DateTimeField(
        verbose_name="Date and time subject taken off schedule",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=get_utcnow,
    )

    def save(self, *args, **kwargs):
        try:
            self._meta.offschedule_datetime_field
        except AttributeError:
            offschedule_datetime_field = "offschedule_datetime"
        else:
            offschedule_datetime_field = self._meta.offschedule_datetime_field
        if not offschedule_datetime_field:
            raise ImproperlyConfigured(
                f"Meta attr 'offschedule_datetime_field' "
                f"cannot be None. See model {self.__class__.__name__}"
            )
        dt = getattr(self, offschedule_datetime_field)
        try:
            dt.date()
        except AttributeError:
            date_not_before_study_start(dt)
            date_not_future(dt)
            self.offschedule_datetime = arrow.Arrow.fromdate(dt).datetime
        else:
            datetime_not_before_study_start(dt)
            datetime_not_future(dt)
            self.offschedule_datetime = dt
        self.report_datetime = self.offschedule_datetime
        super().save(*args, **kwargs)

    @property
    def visit_schedule(self):
        """Returns a visit schedule object."""
        return site_visit_schedules.get_by_offschedule_model(
            offschedule_model=self._meta.label_lower
        )[0]

    @property
    def schedule(self):
        """Returns a schedule object."""
        return site_visit_schedules.get_by_offschedule_model(
            offschedule_model=self._meta.label_lower
        )[1]

    class Meta:
        abstract = True
        offschedule_datetime_field = "offschedule_datetime"
        indexes = [
            models.Index(fields=["id", "subject_identifier", "offschedule_datetime", "site"])
        ]



from datetime import date

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings
from edc_consent import site_consents
from edc_consent.consent import Consent
from edc_constants.constants import FEMALE, MALE
from edc_protocol import Protocol
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow

from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.view_mixins import VisitScheduleViewMixin
from edc_visit_schedule.visit_schedule import VisitSchedule
from visit_schedule_app.models import OnSchedule, SubjectConsent


class MyView(VisitScheduleViewMixin):
    kwargs: dict = {}


class MyViewCurrent(VisitScheduleViewMixin):
    kwargs: dict = {}


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
)
class TestViewMixin(SiteTestCaseMixin, TestCase):
    def setUp(self):
        v1_consent = Consent(
            "visit_schedule_app.subjectconsent",
            version="1",
            start=Protocol().study_open_datetime,
            end=Protocol().study_close_datetime,
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.SubjectOffstudy",
            death_report_model="visit_schedule_app.DeathReport",
        )

        self.schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.OnSchedule",
            offschedule_model="visit_schedule_app.OffSchedule",
            consent_model="visit_schedule_app.subjectconsent",
            appointment_model="edc_appointment.appointment",
        )
        self.schedule3 = Schedule(
            name="schedule_three",
            onschedule_model="visit_schedule_app.OnScheduleThree",
            offschedule_model="visit_schedule_app.OffScheduleThree",
            consent_model="visit_schedule_app.subjectconsent",
            appointment_model="edc_appointment.appointment",
        )

        self.visit_schedule.add_schedule(self.schedule)
        self.visit_schedule.add_schedule(self.schedule3)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(self.visit_schedule)

        self.subject_identifier = "12345"
        self.view = MyView()
        self.view.kwargs = dict(subject_identifier=self.subject_identifier)
        self.view.subject_identifier = self.subject_identifier
        self.view.request = RequestFactory()
        self.view.request.META = {"HTTP_CLIENT_IP": "1.1.1.1"}

        self.view_current = MyViewCurrent()
        self.view_current.kwargs = dict(subject_identifier=self.subject_identifier)
        self.view_current.subject_identifier = self.subject_identifier
        self.view_current.request = RequestFactory()
        self.view_current.request.META = {"HTTP_CLIENT_IP": "1.1.1.1"}

        self.subject_consent = SubjectConsent.objects.create(
            subject_identifier="12345",
            consent_datetime=get_utcnow() - relativedelta(seconds=1),
            dob=date(1995, 1, 1),
            identity="11111",
            confirm_identity="11111",
        )

    def test_context(self):
        context = self.view.get_context_data()
        self.assertIn("visit_schedules", context)
        self.assertIn("onschedule_models", context)

    def test_context_not_on_schedule(self):
        context = self.view.get_context_data()
        self.assertEqual(context.get("visit_schedules"), {})
        self.assertEqual(context.get("onschedule_models"), [])

    def test_context_on_schedule(self):
        obj = OnSchedule.objects.create(subject_identifier=self.subject_identifier)
        context = self.view.get_context_data()
        self.assertEqual(
            context.get("visit_schedules"),
            {self.visit_schedule.name: self.visit_schedule},
        )
        self.assertEqual(context.get("onschedule_models"), [obj])

    def test_context_enrolled_current(self):
        obj = OnSchedule.objects.create(subject_identifier=self.subject_identifier)
        context = self.view_current.get_context_data()
        self.assertEqual(context.get("current_onschedule_model"), obj)
        context.get("current_onschedule_model")



from datetime import datetime

from arrow import Arrow
from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag

from edc_visit_schedule.visit import Visit, VisitCodeError, VisitDateError, WindowPeriod


class TestVisit(TestCase):
    def test_repr(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        self.assertIsNotNone(visit.__repr__())

    def test_name(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        self.assertEqual(visit.name, "1000")

    def test_visit_title(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        self.assertEqual(visit.title, "Visit 1000")
        self.assertEqual(str(visit), "Visit 1000")

    def test_visit_datetime(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        visit.timepoint_datetime = datetime(2001, 12, 1)
        self.assertEqual(
            visit.timepoint_datetime,
            Arrow.fromdatetime(datetime(2001, 12, 1), tzinfo="utc"),
        )

    def test_visit_lower_upper_no_datetime(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        try:
            visit.dates.lower
        except VisitDateError:
            pass
        try:
            visit.dates.upper
        except VisitDateError:
            pass

    @tag("arr")
    def test_visit_lower_upper(self):
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )
        visit.timepoint_datetime = Arrow.fromdatetime(
            datetime(2001, 12, 1), tzinfo="utc"
        ).datetime
        self.assertEqual(
            visit.dates.lower,
            Arrow.fromdatetime(datetime(2001, 12, 1), tzinfo="utc").datetime,
        )
        self.assertEqual(
            visit.dates.upper,
            Arrow.fromdatetime(datetime(2001, 12, 7, 23, 59), tzinfo="utc").datetime,
        )

    def test_window_period_days(self):
        wp = WindowPeriod(rlower=relativedelta(days=0), rupper=relativedelta(days=6))
        dt = Arrow.fromdatetime(datetime(2001, 12, 1), tzinfo="utc").datetime
        self.assertEqual(wp.get_window(dt)[0], dt)
        self.assertEqual(wp.get_window(dt).lower, dt)
        self.assertEqual(
            wp.get_window(dt)[1],
            Arrow.fromdatetime(datetime(2001, 12, 7, 23, 59), tzinfo="utc").datetime,
        )
        self.assertEqual(
            wp.get_window(dt).upper,
            Arrow.fromdatetime(datetime(2001, 12, 7, 23, 59), tzinfo="utc").datetime,
        )

    def test_window_period_weeks(self):
        wp = WindowPeriod(rlower=relativedelta(weeks=1), rupper=relativedelta(weeks=6))
        dt = Arrow.fromdatetime(datetime(2001, 12, 8), tzinfo="utc").datetime
        self.assertEqual(
            wp.get_window(dt).lower,
            Arrow.fromdatetime(datetime(2001, 12, 1), tzinfo="utc").datetime,
        )
        self.assertEqual(
            wp.get_window(dt).upper,
            Arrow.fromdatetime(datetime(2002, 1, 19, 23, 59), tzinfo="utc").datetime,
        )

    def test_good_codes(self):
        try:
            Visit(
                code="1000",
                rbase=relativedelta(days=0),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
                timepoint=1,
            )
        except VisitCodeError as e:
            self.fail(f"VisitError unexpectedly raised. Got {e}")
        try:
            Visit(
                code="1000",
                rbase=relativedelta(days=0),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
                timepoint=1,
            )
        except VisitCodeError as e:
            self.fail(f"VisitError unexpectedly raised. Got {e}")

    def test_no_code(self):
        self.assertRaises(
            VisitCodeError,
            Visit,
            code=None,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )

    def test_bad_code_not_string(self):
        self.assertRaises(
            VisitCodeError,
            Visit,
            code=1,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )

    def test_bad_code_format(self):
        self.assertRaises(
            VisitCodeError,
            Visit,
            code="Aa-1",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            timepoint=1,
        )



from edc_auth.auth_objects import EVERYONE
from edc_auth.site_auths import site_auths
from edc_export.auth_objects import EXPORT

site_auths.update_group(
    "edc_visit_schedule.export_subjectschedulehistory",
    "edc_visit_schedule.export_visitschedule",
    name=EXPORT,
)

site_auths.update_group("edc_visit_schedule.view_subjectschedulehistory", name=EVERYONE)



from dateutil.relativedelta import relativedelta

from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Crf, FormsCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

crfs = FormsCollection(Crf(show_order=1, model="visit_schedule_app.crfone", required=True))

visit0 = Visit(
    code="1000",
    title="Day 1",
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
)

visit1 = Visit(
    code="2000",
    title="Day 2",
    timepoint=1,
    rbase=relativedelta(days=1),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
)

visit2 = Visit(
    code="3000",
    title="Day 3",
    timepoint=2,
    rbase=relativedelta(days=2),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
)

visit3 = Visit(
    code="4000",
    title="Day 4",
    timepoint=3,
    rbase=relativedelta(days=3),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
)

schedule = Schedule(
    name="schedule",
    onschedule_model="visit_schedule_app.onschedule",
    offschedule_model="visit_schedule_app.offschedule",
    appointment_model="edc_appointment.appointment",
    consent_model="visit_schedule_app.subjectconsent",
)

schedule.add_visit(visit0)
schedule.add_visit(visit1)
schedule.add_visit(visit2)
schedule.add_visit(visit3)

visit_schedule = VisitSchedule(
    name="visit_schedule",
    offstudy_model="visit_schedule_app.subjectoffstudy",
    death_report_model="visit_schedule_app.deathreport",
)

visit_schedule.add_schedule(schedule)

# visit_schedule2
schedule2 = Schedule(
    name="schedule2",
    onschedule_model="visit_schedule_app.onschedule2",
    offschedule_model="visit_schedule_app.offschedule2",
    appointment_model="edc_appointment.appointment",
    consent_model="visit_schedule_app.subjectconsent",
)

schedule2.add_visit(visit3)
visit_schedule2 = VisitSchedule(
    name="visit_schedule2",
    offstudy_model="visit_schedule_app.subjectoffstudy2",
    death_report_model="visit_schedule_app.deathreport",
)

visit_schedule2.add_schedule(schedule2)

# visit_schedule5
schedule5 = Schedule(
    name="schedule5",
    onschedule_model="visit_schedule_app.onschedulefive",
    offschedule_model="visit_schedule_app.offschedulefive",
    appointment_model="edc_appointment.appointment",
    consent_model="visit_schedule_app.subjectconsent",
)

schedule5.add_visit(visit0)
visit_schedule5 = VisitSchedule(
    name="visit_schedule5",
    offstudy_model="visit_schedule_app.subjectoffstudyfive",
    death_report_model="visit_schedule_app.deathreport",
)

visit_schedule5.add_schedule(schedule5)

# visit_schedule6
schedule6 = Schedule(
    name="schedule6",
    onschedule_model="visit_schedule_app.onschedulesix",
    offschedule_model="visit_schedule_app.offschedulesix",
    appointment_model="edc_appointment.appointment",
    consent_model="visit_schedule_app.subjectconsent",
)

schedule6.add_visit(visit0)
visit_schedule6 = VisitSchedule(
    name="visit_schedule6",
    offstudy_model="visit_schedule_app.subjectoffstudysix",
    death_report_model="visit_schedule_app.deathreport",
)

visit_schedule6.add_schedule(schedule6)

# visit_schedule7
schedule7 = Schedule(
    name="schedule7",
    onschedule_model="visit_schedule_app.onscheduleseven",
    offschedule_model="visit_schedule_app.offscheduleseven",
    appointment_model="edc_appointment.appointment",
    consent_model="visit_schedule_app.subjectconsent",
)

schedule7.add_visit(visit0)
visit_schedule7 = VisitSchedule(
    name="visit_schedule7",
    offstudy_model="visit_schedule_app.subjectoffstudyseven",
    death_report_model="visit_schedule_app.deathreport",
)

visit_schedule7.add_schedule(schedule7)



from collections import namedtuple

import arrow


class WindowPeriod:
    def __init__(self, rlower=None, rupper=None, visit_code=None):
        self.rlower = rlower
        self.rupper = rupper

    def get_window(self, dt=None):
        """Returns a named tuple of the lower and upper values."""
        dt_floor = arrow.get(dt).to("utc").replace(hour=0, minute=0).datetime
        dt_ceil = arrow.get(dt).to("utc").replace(hour=23, minute=59).datetime
        Window = namedtuple("window", ["lower", "upper"])
        return Window(dt_floor - self.rlower, dt_ceil + self.rupper)



import arrow
from dateutil.relativedelta import relativedelta
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_consent import site_consents
from edc_consent.consent import Consent
from edc_constants.constants import FEMALE, MALE
from edc_facility.import_holidays import import_holidays
from edc_protocol import Protocol
from edc_reference import site_reference_configs
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow
from edc_visit_tracking.constants import SCHEDULED

from edc_visit_schedule.constants import OFF_SCHEDULE, ON_SCHEDULE
from edc_visit_schedule.models import SubjectScheduleHistory
from edc_visit_schedule.site_visit_schedules import (
    RegistryNotLoaded,
    site_visit_schedules,
)
from visit_schedule_app.models import (
    BadOffSchedule1,
    CrfOne,
    OffSchedule,
    OffScheduleFive,
    OffScheduleSeven,
    OffScheduleSix,
    OnSchedule,
    OnScheduleFive,
    OnScheduleSeven,
    OnScheduleSix,
    SubjectConsent,
    SubjectVisit,
)
from visit_schedule_app.visit_schedule import (
    visit_schedule,
    visit_schedule5,
    visit_schedule6,
    visit_schedule7,
)


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
)
class TestModels(SiteTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        import_holidays()
        super().setUpClass()

    def setUp(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "visit_schedule_app.subjectvisit"}
        )
        v1_consent = Consent(
            "visit_schedule_app.subjectconsent",
            version="1",
            start=Protocol().study_open_datetime,
            end=Protocol().study_close_datetime,
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )
        self.subject_identifier = "1234"
        site_consents.registry = {}
        site_consents.register(v1_consent)

    def test_str(self):
        SubjectConsent.objects.create(subject_identifier=self.subject_identifier)
        obj = OnSchedule.objects.create(subject_identifier=self.subject_identifier)
        self.assertIn(self.subject_identifier, str(obj))
        self.assertEqual(obj.natural_key(), (self.subject_identifier,))
        self.assertEqual(
            obj,
            OnSchedule.objects.get_by_natural_key(subject_identifier=self.subject_identifier),
        )

    def test_str_offschedule(self):
        SubjectConsent.objects.create(subject_identifier=self.subject_identifier)
        OnSchedule.objects.create(subject_identifier=self.subject_identifier)
        obj = OffSchedule.objects.create(subject_identifier=self.subject_identifier)
        self.assertIn(self.subject_identifier, str(obj))
        self.assertEqual(obj.natural_key(), (self.subject_identifier,))
        self.assertEqual(
            obj,
            OffSchedule.objects.get_by_natural_key(subject_identifier=self.subject_identifier),
        )

    def test_offschedule_custom_field_datetime(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule5)

        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=2),
        )
        OnScheduleFive.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=2),
        )

        offschedule_datetime = get_utcnow() - relativedelta(years=1)
        obj = OffScheduleFive.objects.create(
            subject_identifier=self.subject_identifier,
            my_offschedule_datetime=offschedule_datetime,
        )
        self.assertEqual(obj.my_offschedule_datetime, offschedule_datetime)
        self.assertEqual(obj.offschedule_datetime, offschedule_datetime)

    def test_offschedule_custom_field_date(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule6)

        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=2),
        )
        OnScheduleSix.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=2),
        )

        offschedule_datetime = arrow.Arrow.fromdate(
            (get_utcnow() - relativedelta(years=1)).date()
        )
        obj = OffScheduleSix.objects.create(
            subject_identifier=self.subject_identifier,
            my_offschedule_date=offschedule_datetime.date(),
        )
        self.assertEqual(obj.my_offschedule_date, offschedule_datetime.date())
        self.assertEqual(obj.offschedule_datetime, offschedule_datetime)

    def test_bad_offschedule1(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule6)

        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=2),
        )
        OnScheduleSix.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=2),
        )

        self.assertRaises(
            ImproperlyConfigured,
            BadOffSchedule1.objects.create,
            subject_identifier=self.subject_identifier,
            my_offschedule_date=get_utcnow(),
        )

    def test_offschedule_no_meta_defaults_offschedule_field(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule7)

        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=2),
        )
        OnScheduleSeven.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=2),
        )

        offschedule_datetime = get_utcnow()
        obj = OffScheduleSeven.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=offschedule_datetime,
        )

        self.assertEqual(obj.offschedule_datetime, offschedule_datetime)

    def test_onschedule(self):
        """Asserts cannot access without site_visit_schedule loaded."""
        site_visit_schedules.loaded = False
        self.assertRaises(
            RegistryNotLoaded,
            OnSchedule.objects.create,
            subject_identifier=self.subject_identifier,
        )

    def test_offschedule_raises(self):
        """Asserts cannot access without site_visit_schedule loaded."""
        site_visit_schedules.loaded = False
        self.assertRaises(
            RegistryNotLoaded,
            OffSchedule.objects.create,
            subject_identifier=self.subject_identifier,
        )

    def test_on_offschedule(self):
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=3),
        )
        OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=3),
        )
        history_obj = SubjectScheduleHistory.objects.get(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(history_obj.schedule_status, ON_SCHEDULE)
        OffSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=get_utcnow(),
        )
        history_obj = SubjectScheduleHistory.objects.get(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(history_obj.schedule_status, OFF_SCHEDULE)

    def test_history(self):
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=3),
        )
        OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=3),
        )
        OffSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=get_utcnow(),
        )
        obj = SubjectScheduleHistory.objects.get(subject_identifier=self.subject_identifier)
        self.assertEqual(
            obj.natural_key(),
            (obj.subject_identifier, obj.visit_schedule_name, obj.schedule_name),
        )
        self.assertEqual(
            SubjectScheduleHistory.objects.get_by_natural_key(
                obj.subject_identifier, obj.visit_schedule_name, obj.schedule_name
            ),
            obj,
        )

    def test_crf(self):
        """Assert can enter a CRF."""
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(years=3),
        )
        OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=get_utcnow() - relativedelta(years=3),
        )
        appointments = Appointment.objects.all()
        self.assertEqual(appointments.count(), 4)
        appointment = Appointment.objects.all().order_by("appt_datetime").first()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            subject_identifier=self.subject_identifier,
            reason=SCHEDULED,
        )
        CrfOne.objects.create(
            subject_visit=subject_visit, report_datetime=appointment.appt_datetime
        )
        OffSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=appointment.appt_datetime,
        )
        self.assertEqual(Appointment.objects.all().count(), 1)

    def test_onschedules_manager(self):
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(months=3),
        )
        onschedule_datetime = get_utcnow() - relativedelta(months=3)

        onschedule = OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=onschedule_datetime,
        )

        history = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual([onschedule], [obj for obj in history])

        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier, report_datetime=get_utcnow()
        )
        self.assertEqual([onschedule], [obj for obj in onschedules])

        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier,
            report_datetime=onschedule_datetime - relativedelta(months=4),
        )
        self.assertEqual(0, len(onschedules))

        # add offschedule
        offschedule_datetime = onschedule_datetime + relativedelta(months=2)
        OffSchedule.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=offschedule_datetime,
        )

        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier,
            report_datetime=offschedule_datetime + relativedelta(days=1),
        )
        self.assertEqual(0, len(onschedules))

        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier,
            report_datetime=offschedule_datetime,
        )
        self.assertEqual([onschedule], [obj for obj in onschedules])

        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier,
            report_datetime=offschedule_datetime - relativedelta(months=1),
        )
        self.assertEqual([onschedule], [obj for obj in onschedules])
        onschedules = SubjectScheduleHistory.objects.onschedules(
            subject_identifier=self.subject_identifier,
            report_datetime=offschedule_datetime + relativedelta(months=1),
        )
        self.assertEqual(0, len(onschedules))

    def test_natural_key(self):
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(months=3),
        )
        obj = OnSchedule.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(obj.natural_key(), (self.subject_identifier,))
        obj = OffSchedule.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(obj.natural_key(), (self.subject_identifier,))



from .constants import DAYS, HOURS, MONTHS, OFF_SCHEDULE, ON_SCHEDULE, WEEKS, YEARS

SCHEDULE_STATUS = ((ON_SCHEDULE, "On schedule"), (OFF_SCHEDULE, "Off schedule"))

VISIT_INTERVAL_UNITS = (
    (HOURS, "Hours"),
    (DAYS, "Days"),
    (WEEKS, "Weeks"),
    (MONTHS, "Months"),
    (YEARS, "Years"),
)



# Generated by Django 2.0 on 2017-12-30 13:43
import _socket
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_model.models
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_protocol.validators
import edc_utils
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SubjectScheduleHistory",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(max_length=50, verbose_name="Subject Identifier"),
                ),
                (
                    "visit_schedule_name",
                    models.CharField(
                        editable=False,
                        help_text='the name of the visit schedule used to find the "schedule"',
                        max_length=25,
                    ),
                ),
                ("schedule_name", models.CharField(editable=False, max_length=25)),
                ("onschedule_model", models.CharField(max_length=100)),
                ("offschedule_model", models.CharField(max_length=100)),
                (
                    "onschedule_datetime",
                    models.DateTimeField(
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.models.datetime_not_future,
                        ]
                    ),
                ),
                (
                    "offschedule_datetime",
                    models.DateTimeField(
                        null=True,
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.models.datetime_not_future,
                        ],
                    ),
                ),
                (
                    "schedule_status",
                    models.CharField(
                        choices=[
                            ("onschedule", "On schedule"),
                            ("offschedule", "Off schedule"),
                        ],
                        max_length=15,
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="subjectschedulehistory",
            unique_together={("subject_identifier", "visit_schedule_name", "schedule_name")},
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils import timezone
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_sites.models import CurrentSiteManager as BaseCurrentSiteManager
from edc_sites.models import SiteModelMixin
from edc_utils import convert_php_dateformat


class CurrentSiteManager(BaseCurrentSiteManager):
    use_in_migrations = True

    def get_by_natural_key(self, subject_identifier):
        return self.get(subject_identifier=subject_identifier)


class ScheduleModelMixin(UniqueSubjectIdentifierFieldMixin, SiteModelMixin, models.Model):
    """A model mixin for a schedule's on/off schedule models."""

    report_datetime = models.DateTimeField(editable=False)

    on_site = CurrentSiteManager()

    objects = SubjectIdentifierManager()

    def __str__(self):
        formatted_date = timezone.localtime(self.report_datetime).strftime(
            convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        )
        return f"{self.subject_identifier} {formatted_date}"

    def natural_key(self):
        return (self.subject_identifier,)

    class Meta:
        abstract = True
